import re

import loopy as lp
import pymbolic.primitives as prim
import islpy as isl

from dune.codegen.generation import (get_counted_variable,
                                     get_global_context_value,
                                     )
from dune.codegen.loopy.transformations.remove_reductions import remove_all_reductions
from dune.codegen.options import get_form_option, get_option
from dune.codegen.pdelab.geometry import world_dimension
from dune.codegen.error import CodegenAutotuneError
from dune.codegen.sumfact.autotune import autotune_realization


def move_zero_assignment_up(kernel, move_up_inames):
    if len(move_up_inames) == 0:
        return kernel

    # Find the instruction we want to move around
    cond = lp.match.Tagged('set_zero')
    instructions = lp.find_instructions(kernel, cond)
    move_iname_set = set(map(lambda x: prim.Variable(x), move_up_inames))
    instr = None
    for i in instructions:
        instr_iname_set = set(i.assignee.index_tuple)
        if move_iname_set.issubset(instr_iname_set):
            # There should be only one matching instruction
            assert instr is None
            instr = i
    assert instr is not None

    # Remove it
    kernel = lp.remove_instructions(kernel, set([instr.id]))

    # Create loop domains: In order to move it upwards we need to create
    # additional loops
    iname_appendix = '_move_up'
    domains = kernel.domains
    for iname in move_up_inames:
        # Find loop bound for this iname
        for dom in domains:
            if iname in dom.get_var_names(isl.dim_type.set):
                # index = dom.get_var_names(isl.dim_type.set).index(iname)

                # TODO: Noch unklar wie man die Loop bound aus isl rausbekommt.
                todo_begin = str(dom).find(iname + ' =') + len(iname) + 3
                if todo_begin == len(iname) + 3 - 1:
                    todo_begin = str(dom).find(iname + ' <=') + len(iname) + 4
                todo_end = todo_begin + str(dom)[todo_begin:].find(' ')
                loop_bound = int(str(dom)[todo_begin:todo_end]) + 1
                break

        domain = "{{ [{0}] : 0<={0}<{1} }}".format(iname + iname_appendix, loop_bound)
        domain = lp.kernel.creation.parse_domains(domain, {})
        domains = domains + domain

    # Create tuple of correct inames for usage in  subscript below
    indices = list(instr.assignee.index_tuple)
    for i in range(len(indices)):
        if indices[i].name in move_up_inames:
            indices[i] = prim.Variable(indices[i].name + iname_appendix)
    indices = tuple(indices)

    # The new instructions needs to lie within those inames
    within_inames = []
    for i in indices:
        within_inames.append(i.name)

    # Create new instruction
    assignee = prim.Subscript(instr.assignee.aggregate, indices)
    instructions = []
    instructions.append(instr.copy(assignee=assignee,
                                   within_inames=frozenset(within_inames)))
    kernel = kernel.copy(instructions=kernel.instructions + instructions,
                         domains=domains)

    # Add dependency to inner assignment instructions
    cond = lp.match.Tagged('assignment')
    assignment_instructions = lp.find_instructions(kernel, cond)
    instr = None
    for i in assignment_instructions:
        instr_iname_set = set(i.assignee.index_tuple)
        if move_iname_set.issubset(instr_iname_set):
            # There should be only one matching instruction
            assert instr is None
            instr = i

    id_zero = instructions[0].id
    cond = lp.match.Id(instr.id)
    kernel = lp.add_dependency(kernel, cond, id_zero)

    return kernel


def _reorder_loops_in_tensor_contraction_direct(kernel, iname_order):
    """Reorder the loop nest of the tensor contractions

    iname_order is a string that specifies the loop order. We use the following convention:

    Each contraction in the sum factorization kernel has the form 'ij,jkl->kli'
    using einsum notation from numpy. iname_order should be a string like
    'iklj' if the loops should be done in order i, k, l, j.

    Without transformations those loops will be done in the order lkij.

    In the sum factorization kernel itself those inames are called:

    sf_out_inames_2_* : l
    sf_out_inames_1_* : k
    sf_out_inames_0_* : i
    red_* : j

    where * represents the current direction (0,1,2 for 3D problems).

    TODO: Maybe also support a list of the inames above?

    TODO: Different order for different direction? Could make sense when we use
    fastdg and a broadcast since the first contraction has a smaller input
    matrix.
    """
    dim = world_dimension()
    # TODO: In principle there is no need to be dimension dependent. I'm just
    # not sure how to pass the iname_order in the general case. This probably
    # needs a rework anyway so I just do the 3D case first.
    assert dim == 3

    kernel = remove_all_reductions(kernel)

    # TODO: Doc after rewrite
    reduction_iname = 'j'
    iname_dict = {'l': 'sf_out_inames_2',
                  'k': 'sf_out_inames_1',
                  'i': 'sf_out_inames_0',
                  'j': 'sf_red'}
    reduction_index = iname_order.index(reduction_iname)
    move_up_inames = list(map(lambda x: iname_dict[x], iname_order[reduction_index + 1:]))

    # cond = lp.match.Tagged('set_zero')
    cond = lp.match.Tagged('assignment')
    instructions = lp.find_instructions(kernel, cond)
    for instr in instructions:
        inames = tuple(map(lambda x: x.name, instr.assignee.index_tuple))
        current_move_up_inames = []
        for i in inames:
            for j in move_up_inames:
                if i.find(j) >= 0:
                    current_move_up_inames.append(i)

        kernel = move_zero_assignment_up(kernel, current_move_up_inames)

        # TODO: There should be a better method than searching the string for
        # 'sf_red'. Unfortunately there are sometimes Call instructions due to
        # broadcasts. That makes different ways difficult.
        regex = re.compile('sf_red_([0-9]*)')
        reduction_index = set(regex.findall(str(instr)))
        assert len(reduction_index) == 1
        reduction_index = reduction_index.pop()
        reduction_iname = 'sf_red_{}'.format(reduction_index)

        prefered_iname_order = []
        for i in inames:
            if i not in current_move_up_inames and i.find('vec') == -1:
                prefered_iname_order.append(i)
        prefered_iname_order.append(reduction_iname)
        for i in current_move_up_inames:
            prefered_iname_order.append(i)
        prefered_iname_order = tuple(prefered_iname_order)
        kernel = lp.prioritize_loops(kernel, prefered_iname_order)

    return kernel


def _reorder_loops_in_tensor_contraction_accum(kernel, iname_order):
    dim = world_dimension()
    assert dim == 3

    if iname_order.endswith('j'):
        return kernel

    # kernel = remove_all_reductions(kernel)
    kernel = _reorder_loops_in_tensor_contraction_direct(kernel, iname_order)

    cond = lp.match.Tagged('set_zero')
    for instr in lp.find_instructions(kernel, cond):
        assert len(instr.depends_on) == 0

        # Depending on this instruction
        depending = []
        for i in kernel.instructions:
            if instr.id in i.depends_on:
                depending.append(i.id)
        assert len(depending) == 1

        active_inames = [i.name.endswith('move_up') for i in instr.assignee.index]
        from loopy.kernel.array import VectorArrayDimTag
        agg = kernel.temporary_variables[instr.assignee.aggregate.name]
        if isinstance(agg.dim_tags[-1], VectorArrayDimTag):
            active_inames[-1] = True

        # Instead of setting the variable itself to zero we set an accumulation
        # variable to zero
        kernel = lp.remove_instructions(kernel, set([instr.id]))
        accum_variable = get_counted_variable('accum_variable')
        accum_init_inames = tuple(i for (i,j) in zip(instr.assignee.index, active_inames) if j)
        assignee = prim.Subscript(prim.Variable(accum_variable,), accum_init_inames)
        accum_init_id = instr.id + '_accum_init'
        accum_init_instr = lp.Assignment(assignee,
                                         0,
                                         within_inames=instr.within_inames,
                                         id=accum_init_id,
                                         tags=('accum_variable',)
                                         )
        kernel = kernel.copy(instructions=kernel.instructions + [accum_init_instr,])

        # Restore dependencies
        for dep in depending:
            match = lp.match.Id(dep)
            kernel = lp.add_dependency(kernel, match, accum_init_id)

        # Make accumulation variable a temporary_variable of this kernel
        #
        # Create dim tags for accum variable
        dim_tags = ','.join(['f'] * sum(active_inames))
        if isinstance(agg.dim_tags[-1], VectorArrayDimTag):
            dim_tags = ','.join(['f'] * (sum(active_inames)-1)) + ",vec"

        # Create shape for accum variable
        shape = tuple(i for (i,j) in zip(agg.shape, active_inames) if j)

        from dune.codegen.loopy.temporary import DuneTemporaryVariable
        var = {accum_variable: DuneTemporaryVariable(accum_variable,
                                                     dtype=agg.dtype,
                                                     shape=shape,
                                                     dim_tags=dim_tags,
                                                     managed=True)}
        kernel.temporary_variables.update(var)

        # Accumulate in accumulate variable
        #
        # Find accumulation instruction
        accum_instr = lp.find_instructions(kernel, lp.match.Id(depending[0]))[0]
        assert accum_instr.assignee == accum_instr.expression.children[0]

        # Dependencies
        depends_on = accum_instr.depends_on
        depending = []
        for i in kernel.instructions:
            if accum_instr.id in i.depends_on:
                depending.append(i.id)

        # Replace with accumulation in accum_variable
        kernel = lp.remove_instructions(kernel, set([accum_instr.id]))
        accum_inames = tuple(i for (i,j) in zip(accum_instr.assignee.index, active_inames) if j)
        assignee = prim.Subscript(prim.Variable(accum_variable,), accum_inames)
        expression = prim.Sum((assignee, accum_instr.expression.children[1]))
        accum_id = accum_instr.id + '_accumvar'
        new_accum_instr = lp.Assignment(assignee,
                                        expression,
                                        within_inames=accum_instr.within_inames,
                                        id=accum_id,
                                        depends_on=depends_on,
                                        )
        kernel = kernel.copy(instructions=kernel.instructions + [new_accum_instr,])

        # Assign accumulation result
        #
        # The reduction is already done
        within_inames = frozenset(i for i in accum_instr.within_inames if 'red' not in i)
        assign_id = accum_instr.id + '_assign'
        assign_instr = lp.Assignment(accum_instr.assignee,
                                     assignee,
                                     within_inames=within_inames,
                                     id=assign_id,
                                     depends_on=frozenset([accum_id,]),
                                     )
        kernel = kernel.copy(instructions=kernel.instructions + [assign_instr,])

        for dep in depending:
            match = lp.match.Id(dep)
            kernel = lp.add_dependency(kernel, match, assign_id)

    return kernel


def reorder_loops_in_tensor_contraction(kernel, iname_order, accum_variable=True):
    if accum_variable:
        return _reorder_loops_in_tensor_contraction_accum(kernel, iname_order)
    else:
        return _reorder_loops_in_tensor_contraction_direct(kernel, iname_order)


def tensor_contraction_loop_order_generator(kernel):
    dim = world_dimension()
    assert dim == 3

    indices = ['l', 'k', 'i', 'j']
    import itertools
    for loop_order in itertools.permutations(indices):
        # palpo TODO: Heavy culling for 'quick' tests during development
        if loop_order[0] != 'l' or loop_order[1] != 'k':
            continue

        order = ''.join(loop_order)
        new_kernel = reorder_loops_in_tensor_contraction(kernel, order, True)
        yield new_kernel, ['reorder_loops_in_tensor_contraction_{}_True'.format(order),]

        new_kernel = reorder_loops_in_tensor_contraction(kernel, loop_order, False)
        yield new_kernel, ['reorder_loops_in_tensor_contraction_{}_False'.format(order),]


def simple_autotuner(kernel_generator, signature):
    # palpo TODO
    from dune.codegen.options import set_option
    set_option("autotune_google_benchmark", True)

    kernel, transformations = next(kernel_generator)
    best_cost = autotune_realization(kernel=kernel, signature=signature, transformations=transformations)
    best_kernel = kernel
    for kernel, transformations in kernel_generator:
        cost = autotune_realization(kernel=kernel, signature=signature, transformations=transformations)
        if cost < best_cost:
            best_cost = cost
            best_kernel = kernel
    return best_kernel


def autotune_tensor_contraction_loop_order(kernel, signature):
    from dune.codegen.loopy.transformations.matchfma import match_fused_multiply_add
    kernel = match_fused_multiply_add(kernel)

    generator = tensor_contraction_loop_order_generator(kernel)
    return simple_autotuner(generator, signature)


def sumfact_performance_transformations(kernel, signature):
    if kernel.name.startswith('sfimpl'):
        # kernel = reorder_loops_in_tensor_contraction_with_accumulation_variable(kernel, "ljik")
        # kernel = autotune_tensor_contraction_loop_order(kernel, signature)

        pass
    return kernel
